unit divtxtdb;

{
    Program for make just text DB.
    For GNU/Linux 64 bit version.
    Written on FreePascal.
    Copyright (C) 2010, 2019 Artyomov N Alexander

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
}

{$MODE OBJFPC}
{$LONGSTRINGS ON}
{$RANGECHECKS ON}
{$SMARTLINK ON}

interface

uses SysUtils{, cwstring};

var
     ds : widestring = '='; // разделитель пар (глобально)

const
     panch = false; pval = true;
type
     TAoL = array of LongInt; // массив длинных целых
     TAoW = array of widestring; // массив юникодовых строк
     TPairStr = array[boolean] of widestring; // пара строк false - якорь, true - данные
     TPair = class(TObject) // пара строк: якорь и данные
     fds : widestring;
     p : TPairStr;
     private
     procedure InString(s : widestring);
     function OutString : widestring;
     function OutValInt : LongInt;   // получить значение как целое число. Если невозможно - поднимается исключение
     procedure InValInt(I : LongInt);
     function OutValFP : extended;   // получить значение как нецелое число. Если невозможно - поднимается исключение
     procedure InValFP(E : extended);
     public
     property Data : widestring read p[pval] write p[pval];     // данные
     property Anchor : widestring read p[panch] write p[panch]; // якорь
     property st : widestring read OutString write InString;    // импорт/экспорт из строки/в строку
     constructor Create;
     constructor Create(d : widestring); // назначить свой разделитель пар
     function IsAEmpty : boolean;    // якорь пуст
     function IsDEmpty : boolean;    // даные пусты
     function IsEmpty : boolean;     // якорь и/или данные пусты
     function IsFEmpty : boolean;    // якорь и данные пусты 
     function IEmpty : LongInt;      // значение пустоты числом (1-все полн,2-якорь пуст,3-данн. пусты,4-все пусты)
     property Int : LongInt read OutValInt write InValInt;
     property FP : extended read OutValFP write InValFP;
     procedure ExCh;                 // обменяться: данным присваивается якорь и наоборот
     end;
     TPairs = class(TObject)         // массив пар строк с разделением одной строки (в т.ч. в файле) табуляцией
     a : array of TPair;
     fds : widestring;
     private
     function ToFile(fn : widestring; save : boolean) : LongInt;
     procedure AppendString(s : widestring);
     procedure InString(s : widestring);
     function OutString : widestring;
     public
     constructor Create;
     constructor Create(d : widestring); // назначить свой разделитель пар
     procedure Clean;
     function TryAddStr(s : widestring) : boolean;
     function AppendFromFile(fn : widestring) : LongInt;
     function LoadFromFile(fn : widestring) : LongInt;
     function AppendToFile(fn : widestring) : LongInt;
     function SaveToFile(fn : widestring) : LongInt;
     property td : widestring read OutString write InString;
     property Appendtd : widestring write AppendString;
     function IsEmpty : boolean;
     procedure ExCh(i1, i2 : LongInt);  // обменяться указанными парами
     procedure FastDelete(i : LongInt); // быстро удалить пару (с нарушением порядка)
     procedure Delete(i : LongInt); // удалить пару
     end;
     TTabbedFile = class(TObject)     // массив массивов пар строк с разделением (в файле) по строкам
     t : array of TPairs;
     private
     function ToFile(fn : widestring; save : boolean) : LongInt;
     public
     procedure Clean;
     function TryAddStr(s : widestring) : boolean;
     function LoadFromFile(fn : widestring) : LongInt;
     function AppendToFile(fn : widestring) : LongInt;
     function AppendFromFile(fn : widestring) : LongInt;
     function SaveToFile(fn : widestring) : LongInt;
     procedure ExCh(i1, i2 : LongInt);  // обменяться указанными массивами пар
     procedure FastDelete(i : LongInt); // быстро удалить строку (с нарушением порядка)
     procedure Delete(i : LongInt); // удалить строку
     end;


implementation

(* ОБЩЕУПОТРЕБИТЕЛЬНЫЕ ФУНКЦИИ *)

procedure RCleanStr(var s : widestring); var f : LongInt;
begin for f := Length(s) downto 1 do if s[f] > #32 then begin SetLength(s, f); Exit; end; s := ''; end;
procedure LCleanStr(var s : widestring); var f, l : LongInt;
begin l := Length(s); for f := 1 to l do if s[f] > #32 then begin s := Copy(s, f, l - f + 1); Exit; end;  s := ''; end;
procedure NeatStr(var s : widestring); // очистка строки
begin RCleanStr(s); LCleanStr(s); end;

(* НАЧАЛО РАБОТЫ С TPAIR *)

function TPair.OutValInt : LongInt; begin try OutValInt := StrToInt(p[pval]); except raise; end; end;
procedure TPair.InValInt(I : LongInt); begin p[pval] := IntToStr(I); end;
function TPair.OutValFP : extended; begin try OutValFP := StrToFloat(p[pval]); except raise; end; end;
procedure TPair.InValFP(E : extended); begin p[pval] := FloatToStr(E); end;

procedure TPair.InString(s : widestring);
  var
      l : LongInt;
begin
l := Pos(fds, s);
p[panch] := Copy(s, 1, l - 1);
p[pval] := Copy(s, l + 1, Length(s) - l);
NeatStr(p[pval]); NeatStr(p[panch]);
end;
function TPair.OutString : widestring;
begin
Exit(p[panch] + fds + p[pval]);
end;
constructor TPair.Create;
begin fds := ds; end;
constructor TPair.Create(d : widestring);
begin fds := d; end;
function TPair.IsAEmpty : boolean; begin IsAEmpty := p[panch] = ''; end;
function TPair.IsDEmpty : boolean; begin IsDEmpty := p[pval] = ''; end;
function TPair.IsEmpty : boolean; begin IsEmpty := IsAEmpty or IsDEmpty; end;
function TPair.IsFEmpty : boolean; begin IsFEmpty := IsAEmpty and IsDEmpty; end;
function TPair.IEmpty : LongInt;
begin IEmpty := 1;
if IsAEmpty then Inc(IEmpty); // 2
if IsFEmpty then Inc(IEmpty,2); // 3 или 4
end;
procedure TPair.ExCh; var tmp : widestring; begin tmp := p[pval]; p[pval] := p[panch]; p[panch] := tmp; end;

(* НАЧАЛО РАБОТЫ С TPAIRS *)

constructor TPairs.Create;
begin fds := ds; end;
constructor TPairs.Create(d : widestring);
begin fds := d; end;
function TPairs.TryAddStr(s : widestring) : boolean;
  var
      pair  : TPair;
begin
  pair := TPair.Create(fds); with pair do begin
  st := s; if not isEmpty then begin
              SetLength(a, Length(a) + 1); a[High(a)] := pair;
              Exit(false);
           end else Free; {end if}
  end; {endw}
Exit(true);
end;
procedure TPairs.Clean; begin SetLength(a,0); end;
function TPairs.AppendFromFile(fn : widestring) : LongInt;
  var
      fp : Text;
      s  : utf8string;
begin
FileMode := 0; // только для чтения
Assign(fp, fn);
{$I-}
ReSet(fp);
{$I+} if IOResult <> 0 then Exit(IOResult);
while not eof(fp) do begin
{$I-}
ReadLn(fp, s);
{$I+} if IOResult <> 0 then Exit(IOResult);
  TryAddStr(UTF8decode(s));
end; {wend}
{$I-}
Close(fp);
{$I+} Exit(IOResult);
end;
function TPairs.LoadFromFile(fn : widestring) : LongInt;
begin Clean; Exit(AppendFromFile(fn)); end;
function TPairs.ToFile(fn : widestring; save : boolean) : LongInt;
  var
      fp : Text;
      f  : LongInt;
begin
FileMode := 1; // только для записи
Assign(fp, fn);
{$I-}
if save then ReWrite(fp) else Append(fp);
{$I+} if IOResult <> 0 then Exit(IOResult);
for f := 0 to High(a) do begin
    {$I-}
     if not a[f].isEmpty then WriteLn(fp, UTF8encode(a[f].st));
    {$I+} if IOResult <> 0 then Exit(IOResult);
end; {f}
{$I-}
Close(fp);
{$I+} Exit(IOResult);
end;
function TPairs.AppendToFile(fn : widestring) : LongInt; begin Exit(ToFile(fn, false)); end;
function TPairs.SaveToFile(fn : widestring) : LongInt; begin Exit(ToFile(fn, true)); end;
procedure TPairs.AppendString(s : widestring);
  var
      f, b, l    : LongInt;
      tmp        : widestring;
begin l := Length(s); b := 0;
for f := 1 to l do begin
if (s[f] = #9) or (f = l) then begin
if f < l then tmp := Copy(s, b+1, f-b-1) else tmp := Copy(s, b+1, f-b);
  TryAddStr(tmp);
b := f;
end; {if}
end; {next f}
end;
procedure TPairs.InString(s : widestring); begin Clean; AppendString(s); end;
function TPairs.OutString : widestring;
  var
      f : LongInt;
begin
OutString := '';
for f := 0 to High(a) do begin
 OutString := OutString + a[f].st;
 if f < High(a) then OutString := OutString + #9;
end; {next f}
end;
function TPairs.IsEmpty : boolean; begin Exit(Length(a) = 0); end;
procedure TPairs.ExCh(i1, i2 : LongInt); var tmp : TPair; begin tmp := a[i1]; a[i1] := a[i2]; a[i2] := tmp; end;
procedure TPairs.FastDelete(i : LongInt); begin ExCh(i, High(a)); SetLength(a, High(a)); end;
procedure TPairs.Delete(i : LongInt);
  var  f : LongInt;
begin
for f := i to High(a) - 1 do a[f] := a[f+1];
SetLength(a, High(a));
end;

(* НАЧАЛО РАБОТЫ С TTABBEDFILE *)

function TTabbedFile.TryAddStr(s : widestring) : boolean;
  var
      pairs  : TPairs;
begin
  pairs := TPairs.Create; with pairs do begin
  td := s; if not isEmpty then begin
              SetLength(t, Length(t) + 1); t[High(t)] := pairs;
              Exit(false);
           end else Free; {end if}
  end; {endw}
Exit(true);
end;
procedure TTabbedFile.Clean; begin SetLength(t,0); end;
function TTabbedFile.AppendFromFile(fn : widestring) : LongInt;
  var
      fp : Text;
      s  : utf8string;
begin
FileMode := 0; // только для чтения
Assign(fp, fn);
{$I-}
ReSet(fp);
{$I+} if IOResult <> 0 then Exit(IOResult);
while not eof(fp) do begin
{$I-}
ReadLn(fp, s);
{$I+} if IOResult <> 0 then Exit(IOResult);
  TryAddStr(UTF8decode(s));
end; {wend}
{$I-}
Close(fp);
{$I+} Exit(IOResult);
end;
function TTabbedFile.LoadFromFile(fn : widestring) : LongInt;
begin Clean; Exit(AppendFromFile(fn)); end;
function TTabbedFile.ToFile(fn : widestring; save : boolean) : LongInt;
  var
      fp : Text;
      f  : LongInt;
begin
FileMode := 1; // только для записи
Assign(fp, fn);
{$I-}
if save then ReWrite(fp) else Append(fp);
{$I+} if IOResult <> 0 then Exit(IOResult);
for f := 0 to High(t) do begin
    {$I-}
     if not t[f].isEmpty then WriteLn(fp, UTF8encode(t[f].td));
    {$I+} if IOResult <> 0 then Exit(IOResult);
end; {f}
{$I-}
Close(fp);
{$I+} Exit(IOResult);
end;
function TTabbedFile.AppendToFile(fn : widestring) : LongInt; begin Exit(ToFile(fn, false)); end;
function TTabbedFile.SaveToFile(fn : widestring) : LongInt; begin Exit(ToFile(fn, true)); end;
procedure TTabbedFile.ExCh(i1, i2 : LongInt); var tmp : TPairs; begin tmp := t[i1]; t[i1] := t[i2]; t[i2] := tmp; end;
procedure TTabbedFile.FastDelete(i : LongInt); begin ExCh(i, High(t)); SetLength(t, High(t)); end;
procedure TTabbedFile.Delete(i : LongInt);
  var  f : LongInt;
begin
for f := i to High(t) - 1 do t[f] := t[f+1];
SetLength(t, High(t));
end;

initialization



finalization


end.